/* config.h.  Generated from config.h.in by configure.  */
/* config.h.in.  Generated from configure.ac by autoheader.  */

/* Define to 1 if you have the <antlr3.h> header file. */
#define HAVE_ANTLR3_H 1

/* Define to 1 if you have the <BCUnit/BCUnit.h> header file. */
/* #undef HAVE_BCUNIT_BCUNIT_H */

/* defined when CU_curses_run_tests is available */
/* #undef HAVE_CU_CURSES */

/* defined when CU_get_suite is available */
/* #undef HAVE_CU_GET_SUITE */

/* Define to 1 if you have the <dlfcn.h> header file. */
#define HAVE_DLFCN_H 1

/* Defined when gnutls api is available */
/* #undef HAVE_GNUTLS */

/* Define to 1 if you have the <inttypes.h> header file. */
#define HAVE_INTTYPES_H 1

/* Define to 1 if you have the `dl' library (-ldl). */
#define HAVE_LIBDL 1

/* Define to 1 if you have the `rt' library (-lrt). */
/* #undef HAVE_CLOCK_GETTIME */

#define HAVE_ZLIB 1

/* Define to 1 if you have the <memory.h> header file. */
#define HAVE_MEMORY_H 1

/* Define to 1 if you have the <stdint.h> header file. */
#define HAVE_STDINT_H 1

/* Define to 1 if you have the <stdlib.h> header file. */
#define HAVE_STDLIB_H 1

/* Define to 1 if you have the <strings.h> header file. */
#define HAVE_STRINGS_H 1

/* Define to 1 if you have the <string.h> header file. */
#define HAVE_STRING_H 1

/* Define to 1 if you have the <sys/stat.h> header file. */
#define HAVE_SYS_STAT_H 1

/* Define to 1 if you have the <sys/types.h> header file. */
#define HAVE_SYS_TYPES_H 1

/* Define to 1 if you have the <unistd.h> header file. */
#define HAVE_UNISTD_H 1

/* Define to the sub-directory in which libtool stores uninstalled libraries.
 */
#define LT_OBJDIR ".libs/"

/* Define to 1 if your C compiler doesn't accept -c and -o together. */
/* #undef NO_MINUS_C_MINUS_O */

/* Name of package */
#define PACKAGE "belle-sip"

/* Define to the address where bug reports for this package should be sent. */
#define PACKAGE_BUGREPORT "jehan.monnier@linphone.org"

/* Define to the full name of this package. */
#define PACKAGE_NAME "belle-sip"

/* Define to the full name and version of this package. */
#define PACKAGE_STRING "belle-sip 0.0.1"

/* Define to the one symbol short name of this package. */
#define PACKAGE_TARNAME "belle-sip"

/* Define to the home page for this package. */
#define PACKAGE_URL ""

/* Define to the version of this package. */
//#define PACKAGE_VERSION "0.0.1"

/* Define to 1 if you have the ANSI C header files. */
#define STDC_HEADERS 1

/* Version number of package */
#define VERSION "0.0.1"

/* Defined when antlr 3.4 api is detected */
#define HAVE_ANTLR_STRING_STREAM_NEW 1

#define ENABLE_SERVER_SOCKETS 1
